
//	Buggalo a simple audio keyer / Bugg for hamradio rigs that lacks CW mode.
//	Copyright (C) 2009  SM0KBW / Bengt Andersson
//
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.



// A Bugg is an autokyer for CW - Morse code transmission. 
// This bugg emulates a IAMBIC mode A keyer.
// IAMBIC A mode is left -> dots, right -> dahs, both a stream of dots and dahs
//
// A vox like function keeps the transmit line low when transmitting in audio keyer mode.
// In normal bugg mode this output is used for keying th transmitter - CW break in



#include <htc.h>
#include "buggalo.h"

#define CLK_CONSTH			0xFD
#define CLK_CONSTL			0x80
#define KEY_TR_DELAY		0x300
#define BUG_TR_MUL		6

#define CW_KEY			0
#define CW_BUGG		1

#define MIN_PITCH	0xB0
#define MAX_PITCH	0xC0
#define MIN_DOT_LENGTH	0x20
#define MAX_DOT_LENGTH	0xF0

#define BUGG_MODE		0
#define KEY_MODE		1

#if defined(_12F683)
#define TONE_OFF	{ TRISIO5 = 1; T0IE = 0;}
#define TONE_ON	{TRISIO5 = 0;  T0IE = 1;}
#else
#define TONE_OFF	{ TRIS5 = 1; T0IE = 0;}
#define TONE_ON	{ TRIS5 = 0; T0IE = 1;}
#endif

word tr_cnt;
word tr_delay;

byte pitch_const;
word d_cnt;
word n_cnt;

byte dot_len;
byte dot_len_lo;
byte dot_len_hi;
byte dot_step;

// State flags to keep track of the bugg
bit l_paddle;
bit r_paddle;
bit do_space;
bit do_dah;
bit do_dih;
bit paddle_chg;
bit bugg_sel;
bit brk_in;


void store_mode (void);
void bugg_mode (void);
void get_paddle_state (void);
void key_mode (void);
void setup (byte mode);
void setup_param (byte *param, byte token, byte pos, byte step);


void cw_out (void);
void send_char (byte token);
void send_store (void);


// Initialize the processor 
void main (void)
{
	GIE = 0;
	ADIE = 0;
	CMIE = 0;
	EEIE = 0;
	GPIE = 0;
	INTE = 0;


	ANSEL = 0;		// No analog pins

#if defined(_12F683)
 	CMCON0 = 0x7;
	IOC = 0x0;		// Interrupts on input pins
	T1GE    = 0;		// No gating
#else
	CMCON = 0x7;
	IOCB = 0x0;		// Interrupts on input pins
	TMR1GE    = 0;	// No gating
#endif

	TRISIO = 0x0F;	// pin 0 , 1, 2, 3 in and 4, 5 out

	GPPU = 0;		// weak pull up on all inputs
	WPU = 0x7;
	T0CS = 0;		// Internal source
	T0SE = 1;		// Positive flank
	PSA	= 0;		// Timer 0 Prescaler / 256
	PS2 = 0;
	PS1 = 1;
	PS0 = 0;

	T1CKPS1   = 0;	// No prescaler
	T1CKPS0   = 0;
	T1OSCEN   = 0;	// No lp oscillator
	T1SYNC    = 1;	// No sync
	TMR1CS    = 0;   // Internal clock
	
	TMR1ON = 1;	//  Activate timer 1


	PEIE = 1;
	GIE = 1;
	
	// Stored pitch value
	pitch_const = eeprom_read (PITCH_NUM);

	// Stored speed values
 	dot_len = eeprom_read (DOT_LEN);
	dot_len_lo = eeprom_read (DOT_LEN_LO);
	dot_len_hi = eeprom_read (DOT_LEN_HI);

	brk_in = 1;
	// Select operation mode

	// Use the bugg as a reditional keyer no vox
	if (GPIO2 == 0)	
		brk_in = 0;

	// Use a traditional CW key Audio key and  Break in (vox) 
	if (GPIO1 == 0)
		key_mode ();


	bugg_mode ();
}

#define  DIT_GEAR_NUM	0xA
#define  DOT_GEAR_NUM	0x6
byte dah_cnt;
byte dit_cnt;


// The actual bugg procedure
void bugg_mode (void)
{

	TMR1IE = 1;

	dot_step = 4;
	send_char (CW_B);

	d_cnt = 5 * dot_len;

	dah_cnt = 0;
	dit_cnt = 0;

RESTART:

//	dot_len is controlling the keying speed
	bugg_sel = 0;



	tr_delay = BUG_TR_MUL * dot_len;


	while (1 == 1)
	{

		// #1 is there any paddle activity
		get_paddle_state ();


		// Decide what to do dependig on CW states and paddle status
		if (!bugg_sel  && d_cnt < dot_len)
		{
			if  (l_paddle  &&  r_paddle) 		// Squeeze mode
			{

				bugg_sel = 1;

				if (do_dah)
				{
					do_dih = 1;
					do_dah = 0;
				}
				else
				if (do_dih)	
				{
					do_dah = 1;
					do_dih = 0;
				}
				else
					do_dah = 1;
			}			
			else
			{
	
				if (r_paddle && !do_dah)	// Right paddle active
				{
					bugg_sel = 1;
					do_dah = 1;
					do_dih = 0;
				}

				if (l_paddle  &&  !do_dih) 	// Left paddle active
				{
					bugg_sel = 1;
					do_dih = 1;
					do_dah = 0;
				}

				if (!l_paddle  && !r_paddle)
				{	
					do_dih = do_dah = 0;
					dit_cnt = dah_cnt = 0;
				}

			}
		}

		if (dot_len > 0xC0)
			dot_step = 8;
		else 
		if (dot_len > 0xA0)
			dot_step = 6;
		else 
		if (dot_len > 0x80)
			dot_step = 4;
		else
			dot_step = 2;

		if (dah_cnt > DOT_GEAR_NUM)
		{
			dot_len -= dot_step;
			if (dot_len < dot_len_hi) 
				dot_len = dot_len_hi;	
			dah_cnt -= 1;

			goto RESTART;
		}
		else
		if (dit_cnt > DIT_GEAR_NUM)
		{
			dot_len + = dot_step;
			if (dot_len > dot_len_lo)
				dot_len = dot_len_lo;	
			dit_cnt -= 2;

			goto RESTART;
		}

		cw_out ();

		// Time to turn off the break in control (vox)
		if (brk_in  &&  GPIO4 == 1  &&  tr_cnt-- == 0)	
			GPIO4 = 0;
	
		// Select another mode?
		if (GPIO0 == 0)
		{
			if (l_paddle)
			{	
				setup (BUGG_MODE);
				goto RESTART;
			}
			else
			if (r_paddle)
			{	
				store_mode ();
				goto RESTART;
			}
			else
				send_store ();
		}
	}
}

#define KEY_TEST_NUM 	8
#define KEY_HITS_LIM 	KEY_TEST_NUM - 1
#define KEY_TEST_TIME	0x8

byte  rpaddle_chg;
byte  lpaddle_chg;

// Convert pin state to logical values
// and debounce the manipulator contacts
void get_paddle_state (void)
{
	byte i, t;

	paddle_chg = 0;
	rpaddle_chg = lpaddle_chg = 0;
	
	for (i = 0; i < KEY_TEST_NUM; i++)
	{
		rpaddle_chg += (GPIO1  ==  r_paddle);
		lpaddle_chg += (GPIO2  ==  l_paddle);
		for (t = KEY_TEST_TIME; t != 0; t--)
		;

	}		

	if (rpaddle_chg > KEY_HITS_LIM)
	{
		r_paddle = (GPIO1 == 0) ? 1 : 0;
		paddle_chg = 1;
	}

	if (lpaddle_chg > KEY_HITS_LIM)
	{
		l_paddle = (GPIO2 == 0) ? 1 : 0;
		paddle_chg = 1;
	}
}


// Works just as a traditional CW beeper - key down gives sound.
// A break in (vox) control is included

void key_mode (void)
{
	TMR1IE = 1;
	d_cnt = 0;

	send_char (CW_K);
	d_cnt = 5 * dot_len;

	while (1 == 1)
	{	
		get_paddle_state();

		// Key down
		if (l_paddle)
		{
			GPIO4 = 1;
			TONE_ON;
			 tr_cnt = KEY_TR_DELAY;
		}
		// Key up
		else
			TONE_OFF;

		// Turn off break in?
		if (GPIO4 == 1  &&  tr_cnt-- == 0)	
			GPIO4 = 0;

		// Select another mode?
		if (GPIO0 == 0)
			send_store ();

	}

}


bit lvl;
void interrupt isr (void)
{
	GIE = 0;

	// Tone generator
	// Need to be first in the ISR to avoid extra noise
	if (T0IE == 1  &&  T0IF == 1)
	{
		T0IF = 0;
		TMR0 = pitch_const;
		
		GPIO5 = (GPIO5 == 0) ? 1 : 0;
			
	}


	// Token timer
	if (TMR1IE == 1  &&  TMR1IF == 1)
	{
		TMR1IF = 0;
		TMR1H = CLK_CONSTH;
		TMR1L = CLK_CONSTL;

		if (d_cnt != 0)
			d_cnt--;

	}

	GIE = 1;
}


// Hardware layer control of the  state 
void cw_out (void)
{

	// CW output
	if (d_cnt == 0)
	{
		if (do_space)
		{
			d_cnt = dot_len;
			TONE_OFF;

			if (!brk_in)
				GPIO4 = 0;

		 	tr_cnt = tr_delay;
			do_space = 0;

		}
		else
		if (do_dih)
		{
			d_cnt = dot_len;
			TONE_ON;
			GPIO4 = 1;
			do_space = 1;
			bugg_sel = 0;

			if (brk_in)
		 		tr_cnt = tr_delay;

			dit_cnt++;

		}
		else
		if (do_dah)
		{
			d_cnt = 3 * dot_len;
			TONE_ON;
			GPIO4 = 1;
			do_space = 1;
			bugg_sel = 0;

			if (brk_in)
		 		tr_cnt = tr_delay;

			dah_cnt++;
			
		}
	}
}





#define P_CNT	32
#define P_LIM    10
bit w_spc;


// Procedure that recieves a string of tokens from the paddles and store it in eeproom.
//
// Programming is done by slap keying.
//
// left paddle stores a short, left a long
// Both paddles first time save the ongoing token and a space, thereafter only an extra space.
// Sound marks that a change is made.

void store_mode (void)
{
	byte i;
	byte j;
	byte cw_t;
	byte rem;
	byte lng;
	byte spc_dat;
	byte w_spc;
	byte b_cnt;
	byte s_cnt;

	byte lp;
	byte rp;

	word t;

	PEIE = 0;

	i = CH_STORE;
	j = SPC_STORE;
	cw_t = 0;
	b_cnt = 0;
	spc_dat = 0;
	s_cnt = 0;
		

	// Mark the memory as empty
	eeprom_write(i, CW_END);
	
	while (i < CH_STORE_SZ)
	{

		
		if (T0IE)
		{
			// Beep delay
			for (; t != 0; t--)
			;
			TONE_OFF;
		}

		// Wait until something happens
		paddle_chg = 0;

		while (!paddle_chg)
		{	
			for (t = 0x10; t != 0; t--)
			;

			get_paddle_state ();
		}
		


		rp = lp = 0;
		// Lets sample the paddle state and slow down things a bit
		while (P_CNT != rp  && P_CNT != lp)
		{	
			get_paddle_state ();
			if (l_paddle)
				lp++;

			if (r_paddle)
				rp++;
			
			for (t = 0x180; t != 0; t--)
			;
		}


		// Booth paddle or time to save
		if ((rp > P_LIM  &&  lp > P_LIM)  ||  b_cnt == 5)
		{
			// let there be beep
			TONE_ON;

			// There is a token to save
			if (b_cnt != 0)
			{
				cw_t <<= (7 - b_cnt);

				// A token takes more than one byte to store
				if (b_cnt == 5  &&  (rp == 0  ||  lp == 0))
				{	
					lng = 1;
					b_cnt = 7;
					rem = (rp) ? 2 : 0;
				}
				else
					lng = 0;


				cw_t |= b_cnt;
				
				// add a space
				spc_dat <<= 1;
				s_cnt++;
				b_cnt = 0;
			}
			else
			{
				// Just add a space
				spc_dat |= 2;

				if (s_cnt == 0)
				{
					s_cnt = 8;
					j--;
				}
			}

			// time to save space data
			if (s_cnt == 8)
			{
				s_cnt = 0;
				spc_dat = 0;
				eeprom_write(j++, spc_dat);
			}

			// Time to save token to eeprom and move end mark one step. save space		
			if (cw_t != 0)
			{
				eeprom_write(i++, cw_t);
				eeprom_write(i, CW_END);
				eeprom_write(j, spc_dat);
				
				t = 0xC000;				
			}
			else
				t = 0x5000;	
				
			cw_t = 0;	

			if (lng)
			{	
				b_cnt = 1;
				cw_t = rem;
				t = 0x5000;
				lng = 0;
			}
		}
		else
		if (r_paddle)
		{	// A dah is noted

			// let there be beep
			TONE_ON;

			cw_t |= 1;
			cw_t <<= 1;
			b_cnt++;
			w_spc = 0;
			t = 0x6000;
		}
		else
		if (l_paddle)
		{	// A dih is noted

			// let there be beep
			TONE_ON;

			cw_t <<= 1;
			b_cnt++;
			w_spc = 0;
			t = 0x2000;
		}
	}
	PEIE = 1;
}


// First, if in  bugg_mode, send a 'S' and allow bugg speed change
// Second Send a 'P' and allow pitch change
// third Send a 'C' and exit
void setup (byte bugg_mode)
{
	byte step;

	GPIO4 = 1;
	
	TMR1IE = 1;
	d_cnt = 0;
	

	if (bugg_mode == BUGG_MODE)
	{


		setup_param (&dot_len, CW_S, DOT_LEN, 5);
		setup_param (&dot_len, CW_U, DOT_LEN_HI, 5);
		setup_param (&dot_len, CW_L, DOT_LEN_LO, 5);
		setup_param (&pitch_const, CW_P, PITCH_NUM, 1);

	}

	d_cnt = 10 * dot_len;

	send_char (CW_R);
	d_cnt = 10 * dot_len;

	GPIO4 = 1;
}

void setup_param (byte *param, byte token, byte pos, byte step)
{
		d_cnt = 10 * dot_len;
		send_char (token);
		d_cnt = 10 * dot_len;

		while (GPIO0 == 1)
		{
			get_paddle_state ();

			if (l_paddle && (*param + step) < MAX_DOT_LENGTH)
				*param += step;
			else
			if (r_paddle && (*param - step) > MIN_DOT_LENGTH)
				*param -= step;
		
			send_char (CW_V);
			d_cnt = 3 * dot_len;
		}
		// Save the new speed
		eeprom_write (pos, *param);
		dot_len = eeprom_read (DOT_LEN);
}



// Procedure that convert a stored char to actual keying sequences
void send_char (byte token)
{

	byte i;
	byte t;


	t = token;
	i = t & 0x7;
		

	for (; i > 0; i--)
	{
		// finnish current output
		while ( do_space != 0  || d_cnt  != 0)
			cw_out ();

		// get next output
		if (t & 0x80)
		{
			do_dah = 1;
		}
		else
		{
			do_dih = 1;
		}
	
		cw_out ();
		do_dih = 0;
		do_dah = 0;

		t = t << 1;
	}
	// clean up and finnish

	do_dah = 0; 
	do_dih = 0;

	while (do_space != 0)
		cw_out ();
}

bit space;

// Procedure that reads CW-tokens from eeprom and prepare them for send_char 
void send_store (void)
{
	byte i;
	byte j;
	byte cw_t;
	byte spc_dat;
	byte b_cnt;
	byte s_cnt;

	i = CH_STORE;
	j = SPC_STORE;
	spc_dat = eeprom_read (j++);
	s_cnt = 0;

	cw_t = eeprom_read (i++);

	while (i < CH_STORE_SZ  &&  cw_t != CW_END)
	{
		// Does this token continous in next pos ?
		if ((cw_t & CW_LNK) == CW_LNK)
		{
			space = 0;
			cw_t &= 0xFD;
		}
		else
			space = 1;

		send_char (cw_t);
		
		if (space)
		{
			if ((spc_dat & 0x80)  ==  0x80)
				d_cnt = 7 * dot_len;
			else
				d_cnt = 3 * dot_len;
			
			spc_dat <<= 1;
			s_cnt++;

			if (s_cnt == 8)
			{
				s_cnt = 0;
				spc_dat = eeprom_read (j++);
			}
		}

		
		cw_t = eeprom_read (i++);

	}
	
	dit_cnt = dah_cnt = 0;
}
